;; This file converts obj/opdesc into bin/instructions.html

(defmacro taga (tag attributes &body body)
  "Format a HTML tag with attributes.  Each expression in the body becomes a member of the tag."
  ``("<" ,',tag ,,(when attributes `(mapcar (lambda (a) (list " " (car a) "=" (format nil "~s" (second a))))
                                             ,attributes))
         ,,@(if body (list ">" (cons 'list body) "</" `(quote ,tag) ">") (list "/>"))))

(defmacro tag (tag &optional &body body)
 "Plain HTML tag without attributes"
 `(taga ,tag () ,@body))

(defvar *legend*
  '(("r0-r15" "The 16 registers currently visible in the register window.  r0 is always the head of the register stack.")
    ("rD" "The primary data register that will be affected in the instruction.")
    ("rA" "Auxiliary/address register for 2-reg instructions.  This register tends to be read-only.")
    ("rP" "The prior register, which is rD from the last instruction that used one.  Determines branch tests and is the target of some implied-mode instructions.")
    ("rH" "The register immediately after rD, usually used as the high word of a 32-bit value starting in rD.  For example, if an instruction is using r3 as rD, then rH is r4.  There is no wraparound, so if r15 is used, rH will be the currently unaddressable effective r16.")
    ("rD:rH" "32-bit effective register, with rD being the low word and rH being the high word.")
    ("imm4" "4-bit immediate value, usually 0-15, though some instructions disallow 0.")
    ("imm4p1" "4 bit immediate value interpreted as 1-16.")
    ("imm8" "8-bit immediate value, usually 0-255, or -128 to +127 signed.")
    ("imm16" "16-bit immediate value, usually 0-65535, or -32768 to +32767 signed.")
    ("rel8" "Used in branches, a target address which is within an 8-bit signed range of bytes from the first byte of the instruction.  A branch to a rel8-encoded $00 loops back to the branch instruction itself.")
    (":=" "Assignment")
    ("*" "Embedded instruction, taking 16 opcodes.")))

(defmacro sym-if (var val &body then-else)
  `(let ((,var ,val))
     (if ,var ,@then-else)))

(defmacro sym-when (var val &body body)
  `(let ((,var ,val))
     (when ,var ,@body)))

(defun field (name row)
  (second (assoc name (cdr row))))

(defmacro if-field (name row then &optional else)
  `(let ((,name (field ',name ,row)))
     (if ,name ,then ,else)))

(defun render (item &optional (str t))
  "Unfolds an iolist-style combination of lists and strings to a stream, supporting calling tag operations"
  (when item
    (if (listp item)
        (case (first item)
          ((tag taga) (eval item))
          (otherwise (mapc (lambda (inner) (render inner str)) item)))
        (format str "~a" item))))




;; Main routine
(let ((rows (with-open-file (in "obj/opdesc")
              (read in))))
  (labels ((filter (&rest tags)
             (remove-if (lambda (row) (not (subsetp tags row :test #'equal))) rows)))
    ;; Tack on categories that are not declared
    (mapc (lambda (row)
            (sym-when cat (field 'category row)
              (unless (filter 'category `(name ,cat))
                (nconc rows `((category (name ,cat) (title, cat)))))))
          rows)

    ;; Smash together HTML tags
    (render (tag body
              (tag h1 "Acheron VM - Instruction Set")
              ;; Table of contents
              (tag h2 "Contents")
              (tag ul
                (let ((index 0))
                  (mapcar (lambda (category)
                            (prog1 (tag li (taga a `((href ,(format nil "#~a" index))) (field 'title category)))
                              (incf index)))
                          (cons '(category (title "Naming Conventions")) (filter 'category)))))
              ;; Header
              (tag hr)
              (taga a '((name "0")) (tag h2 "Naming Conventions"))
              (tag blockquote
                (tag table
                  (mapcar (lambda (row) (list (tag tr
                                                (tag td (tag code (tag b (first row))))
                                                (taga td '((width 20)))
                                                (tag td (second row)))))
                          *legend*)))
              (tag p (tag i "The following was auto-generated by AcheronVM build tools") (tag br))
              ;; Per category
              (let ((index 0))
                (mapcar (lambda (category)
                          (let* ((catname (field 'name category))
                                 (catfield (list 'category catname)))
                            (list
                             ;; Title & header
                             (tag hr)
                             (taga a `((name ,(format nil "~a" (incf index)))))
                             (tag h2 (field 'title category))
                             (if-field header category (tag p header))
                             (tag blockquote
                               ;; Instructions, sorted by name
                               (sym-when ops (sort (filter 'op catfield) #'string<
                                                   :key (lambda (x) (field 'name x)))
                                 (tag table
                                   (mapcar (lambda (op)
                                             (tag tr
                                               (tag td (tag b (tag pre (field 'name op))))
                                               (tag td (tag i (tag pre
                                                                (if (eq 'e (field 'embed op)) "*" " ")
                                                                (field 'paramtext op))))
                                               (taga td '((width 20)))
                                               (tag td (field 'doc op))))
                                           ops)))
                               ;; Pseudo instructions
                               (sym-when pseudos (filter 'pseudo-op catfield)
                                 (tag p (tag b "Pseudo Instructions")
                                      (tag table
                                        (mapcar (lambda (p)
                                                  (tag tr
                                                    (tag td (tag b (tag code (field 'name p))))
                                                    (tag td (tag i (tag code (field 'params p))))
                                                    (taga td '((width 10)))
                                                    (tag td (field 'doc p))))
                                                pseudos))))
                               ;; Natives
                               (sym-when natives (filter 'native catfield)
                                 (tag p (tag b "Native Routines")
                                      (tag table
                                        (mapcar (lambda (n)
                                                  (tag tr
                                                    (tag td (tag pre (field 'calltype n) " "
                                                                 (tag b (field 'name n))))
                                                    (taga td '((width 20)))
                                                    (tag td (field 'doc n))
                                                    ))
                                                natives))))
                               ;; Zeropage
                               (sym-when zps (filter 'zeropage catfield)
                                 (tag p (tag b "Zeropage Locations")
                                      (tag table
                                        (mapcar (lambda (zp)
                                                  (tag tr
                                                    (tag td (tag b (tag code (field 'name zp) ":")))
                                                    (tag td (tag pre " .res " (field 'size zp)))
                                                    (taga td '((width 20)))
                                                    (tag td (field 'doc zp))))
                                                zps))))
                               )
                             ;; Documentation
                             (mapcar (lambda (doc) (tag p (field 'doc doc))) (filter 'category-doc catfield))
                             )))
                        (filter 'category)))))
    ))
